#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/XF86keysym.h>
#include <X11/XKBlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/extensions/Xinerama.h>


#define INTERSECT(x,y,w,h,r)  (MAX(0, MIN((x)+(w),(r).x_org+(r).width)  - MAX((x),(r).x_org)) \
                             * MAX(0, MIN((y)+(h),(r).y_org+(r).height) - MAX((y),(r).y_org)))
#define MIN(a,b)              ((a) < (b) ? (a) : (b))
#define MAX(a,b)              ((a) > (b) ? (a) : (b))
#define TTWM_FULLSCREEN	0x0004

enum { Occupied, Selected, LASTColor };
enum { MOff, MMove, MResize };

typedef struct {
	unsigned int mod;
	KeySym keysym;
	void (*func)(const char *);
	const char *arg;
} Key;

typedef struct {
	unsigned int mod, button;
	void (*func)(const char *);
	const char *arg;
} Button;

typedef struct Client Client;
struct Client {
	Window win, parent;
	Client *next;
	int x,y,w,h;
	int tags, flags, screen;
};

 typedef struct {
 const char *instance,*title;
 int workspace, maximized, screen;
} Rule;
/*********************** [1] PROTOTYPES & VARIABLES ***********************/
/* 1.0 EVENT HANDLER PROTOTYPES */
static void buttonpress(XEvent *);
static void buttonrelease(XEvent *);
static void configurerequest(XEvent *);
static void enternotify(XEvent *);
static void expose(XEvent *);
static void keypress(XEvent *);
static void maprequest(XEvent *);
static void motionnotify(XEvent *);
//~ static void propertynotify(XEvent *);
static void unmapnotify(XEvent *);
static void clientmessage(XEvent *);

/* 1.1 BINDABLE FUNCTION PROTOTYPES */
static void fullscreen(const char *);
static void killclient(const char *);
static void mouse(const char *);
static void quit();
static void spawn(const char *);
void screen(const char *arg);
static void tag(const char *);
static void window(const char *);
/* 1.2 TTWM INTERNAL PROTOTYPES */
static int draw();
static int neighbors(Client *);
static GC setcolor(int);
static Client *wintoclient(Window);

/* 1.3 GLOBAL VARIABLES */
#include "config.h"

static Display *dpy;
static Window root;
static int scr, sw, sh;
static Colormap cmap;
static XColor color;
static GC gc,bgc;
static Bool running = True;
static XButtonEvent mouseEvent;
static int mouseMode = MOff;
static Client *clients = NULL, *focused = NULL, *nextwin, *prevwin;
static int tagsSel = 1;
int retval;
static void (*handler[LASTEvent]) (XEvent *) = {
	[ButtonPress]		= buttonpress,
	[ButtonRelease]		= buttonrelease,
	[ConfigureRequest]	= configurerequest,
	[EnterNotify]		= enternotify,
	[Expose]			= expose,
	[KeyPress]			= keypress,
	[MapRequest]		= maprequest,
	[MotionNotify]		= motionnotify,
	[UnmapNotify]		= unmapnotify,
	[ClientMessage]		= clientmessage,
};

/************************ [2] FUNCTION DEFINITIONS ************************/
/* 2.0 EVENT HANDLERS */

void buttonpress(XEvent *ev) {
	Client *c;
	XButtonEvent *e = &ev->xbutton;
	if (!( (c=wintoclient(e->subwindow)) && e->state) ) return;
	if (c && e->button < 4) { focused = c; XRaiseWindow(dpy,c->win); }
	int i;
	mouseEvent = *e;
	for (i = 0; i < sizeof(buttons)/sizeof(buttons[0]); i++)
		if ( (e->button == buttons[i].button) && buttons[i].func &&
				buttons[i].mod == ((e->state&~Mod2Mask)&~LockMask) )
			buttons[i].func(buttons[i].arg);
	if (mouseMode != MOff)
		XGrabPointer(dpy,root,True,PointerMotionMask | ButtonReleaseMask,
				GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
}

void buttonrelease(XEvent *ev) {
	XUngrabPointer(dpy, CurrentTime);
	mouseMode = MOff;
}

void configurerequest(XEvent *ev) {
	XConfigureRequestEvent *e = &ev->xconfigurerequest;
	Client *c;

	if ( (c=wintoclient(e->window)) && (e->width==sw) && (e->height==sh) ) {
		c->flags |= TTWM_FULLSCREEN;
		draw();
		return;
	}
	XWindowChanges wc;
	wc.x = e->x; wc.y = e->y; 
	wc.width = e->width; wc.height = e->height;
	wc.border_width = borderwidth;
	wc.sibling = e->above; wc.stack_mode = e->detail;
	XConfigureWindow(dpy,e->window,e->value_mask,&wc);
	XFlush(dpy);
}

void enternotify(XEvent *ev) {
	if (!focusfollowmouse) return;
	Client *c = wintoclient(ev->xcrossing.window);
	if (c) {
		focused = c;
		draw();
	}
}
int cycle_tags(int i){
	int newtag;
	newtag=tagsSel+i;
	if(newtag==3 && i==1)return 4;
	if(newtag>=5 && i==1)return 1;
	if(newtag==3 && i==-1)return 2;
	if(newtag<1 && i==-1)return 4;
	else return newtag;
}

int checkrules(const char *arg) {
	if (!focused) return 0;
	int i,n;
	char *name;
	XFetchName(dpy,focused->win,&name);
	XClassHint ch = { NULL, NULL };
	XGetClassHint(dpy, focused->win, &ch);
	for (i=0; i < sizeof(rules)/sizeof(rules[0]); i++)
	if (strstr(ch.res_name, rules[i].instance) && (!rules[i].title || strstr(name, rules[i].title))){
		 n++;
		 focused->tags =  rules[i].workspace;
		 tagsSel = cycle_tags(rules[i].workspace);
		 if( rules[i].maximized == 1 ) focused->flags =TTWM_FULLSCREEN;
	 }
	if(ch.res_class) XFree(ch.res_class);
	if(ch.res_name)  XFree(ch.res_name);
	return n;
}

void expose(XEvent *ev) {
	draw();
}

void keypress(XEvent *ev) {
	unsigned int i;
	XKeyEvent *e = &ev->xkey;
	KeySym keysym = XkbKeycodeToKeysym(dpy,(KeyCode)e->keycode,0,0);
	for (i = 0; i < sizeof(keys)/sizeof(keys[0]); i++) {
		if ( (keysym == keys[i].keysym) && keys[i].func &&
				keys[i].mod == ((e->state&~Mod2Mask)&~LockMask) )
			keys[i].func(keys[i].arg);
	}
}

void maprequest(XEvent *ev) {
	int n,x,y, di, i = 0;
	unsigned int du;
	Window dw;
	XineramaScreenInfo *info;
	info = XineramaQueryScreens(dpy, &n);
	XQueryPointer(dpy, root, &dw, &dw, &x, &y, &di, &di, &du);
	for(i = 0; i < n; i++)
		if(INTERSECT(x, y, 1, 1, info[i]))
		break;


	Client *c;
	XWindowAttributes wa;
	XMapRequestEvent *e = &ev->xmaprequest;

	if (!XGetWindowAttributes(dpy,e->window,&wa) || wa.override_redirect) return;
	if (!wintoclient(e->window)) {
	if (!(c=calloc(1,sizeof(Client)))) return;
	
	int j;
	for(j = 0; j <= i; j++){
		if ( (wa.width==info[j].width) && (wa.height==info[j].height) ){
			c->flags ^= TTWM_FULLSCREEN;
			c->screen=j;
		}
	}
	if(!c->screen)c->screen=i;

	c->tags = tagsSel;
	c->parent = e->parent;
	c->win = e->window;
	c->w = wa.width; c->h = wa.height; 
	c->x = info[i].x_org+info[i].width/2-c->w/2;
	c->y = info[i].y_org+info[i].height/2-c->h/2;

	if (c->x < 0) c->x = 0;
	if (c->y < 0) c->y = 0;

	XSelectInput(dpy,c->win,PropertyChangeMask | EnterWindowMask);
	c->next = clients; clients = c;
	XSetWindowBorderWidth(dpy,c->win,borderwidth);
	XMapWindow(dpy,c->win);
	XRaiseWindow(dpy,c->win);
	focused = c;
	if (!wa.bit_gravity==ForgetGravity)checkrules(NULL);
	}
	XFree(info);
	draw();
}

void motionnotify(XEvent *ev) {
	XineramaScreenInfo *info;
	int n;
	info = XineramaQueryScreens(dpy, &n);

	int xdiff, ydiff;
	while (XCheckTypedEvent(dpy,MotionNotify,ev));
	XButtonEvent *e = &ev->xbutton;
	xdiff = e->x_root - mouseEvent.x_root;
	ydiff = e->y_root - mouseEvent.y_root;
	if(mouseMode != MOff && focused->flags & TTWM_FULLSCREEN)
		focused->flags ^=TTWM_FULLSCREEN;
	if (mouseMode == MMove) {
		focused->x+=xdiff; focused->y+=ydiff;
/* stuck window*/
		//~ if(focused->x <0)focused->x=0;
		//~ else if(focused->x >info[focused->screen].width-focused->w)focused->x=info[focused->screen].width-focused->w;
		//~ if(focused->y <0){focused->y=0;draw();}
		//~ else if(focused->y >info[focused->screen].height-focused->h){focused->y=info[focused->screen].height-focused->h;draw();}
/*switch tags*/
		if(focused->x <info[focused->screen].x_org-focused->w/2){
			focused->tags=tagsSel=cycle_tags(-1);
			focused->x=info[focused->screen].width-focused->w/2;
			}
		else if(focused->x > info[focused->screen].width-focused->w/2){
			focused->tags=tagsSel=cycle_tags(1);
			focused->x=0;
			draw();
			}
		if(n>2){
		if(focused->y-focused->h/2 > info[n-1].width-focused->h/2){
			focused->y=0;
			draw();
		}
		else if(focused->y+focused->h/2 < 0){
			focused->y=info[n-1].width-focused->h/2;
			draw();
		}
		}
/*screens*/
		//~ if(focused->y >= info[0].y_org)focused->screen=0;
		//~ else focused->screen=1;
		/*ou*/
		//~ if(focused->y >= info[1].y_org)focused->screen=1;
		//~ else focused->screen=0;
/* */
	}
	else if (mouseMode == MResize) {
		focused->w+=xdiff; focused->h+=ydiff;
	}
	draw();
	mouseEvent.x_root+=xdiff; mouseEvent.y_root+=ydiff;
}

void clientmessage(XEvent *ev) {
	XClientMessageEvent *cme = &ev->xclient;
	Client *c = wintoclient(cme->window);
	if(cme->data.l[0] == IconicState)XUnmapWindow(dpy,c->win);
}
	
void unmapnotify(XEvent *ev) {
	Client *c,*t;
	if (!(c=wintoclient(ev->xunmap.window)) || ev->xunmap.send_event) return;
	if (c == focused) {
		neighbors(c);
		if (nextwin) focused = nextwin;
		else focused = prevwin;
	}
	if (c == clients) clients = c->next;
	else {
		for (t = clients; t && t->next != c; t = t->next);
		t->next = c->next;
	}
	free(c); c = NULL;
	draw();
}

/* 2.1 BINDABLE FUNCTIONS */

void fullscreen(const char *arg) {
	if (!focused) return;
	focused->flags ^= TTWM_FULLSCREEN;
	XRaiseWindow(dpy,focused->win);
	draw();
}

void killclient(const char *arg) {
	if (!focused) return;
	XEvent ev;
	ev.type = ClientMessage; ev.xclient.window = focused->win;
	ev.xclient.message_type = XInternAtom(dpy,"WM_PROTOCOLS",True);
	ev.xclient.format = 32;
	ev.xclient.data.l[0] = XInternAtom(dpy,"WM_DELETE_WINDOW",True);
	ev.xclient.data.l[1] = CurrentTime;
	XSendEvent(dpy,focused->win,False,NoEventMask,&ev);
}

void mouse(const char *arg) {
	if (arg[0] == 'm') mouseMode = MMove;
	else if (arg[0] == 'r') mouseMode = MResize;
}

void quit(const char *arg) {
	running = False;
	retval= arg[0] - 255 -49 ;
}

void spawn(const char *arg) {
	system(arg);
}



void tag(const char *arg) {
	int i = arg[2] - 49;
	if (arg[0] == 's'){
	tagsSel=cycle_tags(i);
}
	else if (arg[0] == 'a' && focused ) focused->tags = cycle_tags(i);
	else if (arg[0] == 'm' && focused ) focused->tags = tagsSel = cycle_tags(i);
	draw();
}

void window(const char *arg) {
	if (!focused || !clients) return;
	neighbors(focused);
	Client *t = NULL;
	if (arg[0] == 'p') t = prevwin;
	else if (arg[0] == 'n') t = nextwin;
	if (!t) return;
	focused = t;
	XRaiseWindow(dpy,focused->win);
	draw();
}

void screen(const char *arg){
	XineramaScreenInfo *info;
	if (arg[0] == 'p')focused->screen=focused->screen-1;
	if (arg[0] == 'n')focused->screen=focused->screen+1;
	int n;
	info = XineramaQueryScreens(dpy, &n);
	if(focused->screen <0)focused->screen=n-1;
	if(focused->screen >=n)focused->screen=0;
	focused->x=info[focused->screen].x_org+(info[focused->screen].width-focused->w)/2;
	focused->y=info[focused->screen].y_org+(info[focused->screen].height-focused->h)/2;
	draw();
	XFree(info);
}
/* 2.2 WM INTERNAL FUNCTIONS */

int draw() {
	int n,x,y, di, i = 0;
		unsigned int du;
		Window dw;
	XineramaScreenInfo *info;
	info = XineramaQueryScreens(dpy, &n);
	XQueryPointer(dpy, root, &dw, &dw, &x, &y, &di, &di, &du);
	for(i = 0; i < n; i++)
		if(INTERSECT(x, y, 1, 1, info[i]))
		break;

	/* windows */
	Client *stack = clients;
	XSetWindowAttributes wa;
	while (stack) {
		if (!(stack->tags & tagsSel)) { /* not on selected tag(s) */
			XMoveWindow(dpy,stack->win,-2*sw,0);
			stack = stack->next;
			continue;			
		}
		else {
				if (stack->flags & TTWM_FULLSCREEN ){
			XMoveResizeWindow(dpy,stack->win,info[stack->screen].x_org,info[stack->screen].y_org,info[stack->screen].width,info[stack->screen].height);
			XSetWindowBorderWidth(dpy,stack->win,0);
		}
		else{
			//~ if(stack->w > info[stack->screen].width) {stack->x=info[stack->screen].x_org;stack->w=info[stack->screen].width;}
			//~ if(stack->h > info[stack->screen].height) {stack->y=info[stack->screen].y_org;stack->h=info[stack->screen].height;}
			XMoveResizeWindow(dpy,stack->win,stack->x,stack->y,stack->w,stack->h);
			XSetWindowBorderWidth(dpy,stack->win,1);
		}
		if (!focused || !(focused->tags & tagsSel)) {
			if (stack->tags == tagsSel) focused = stack;
		}

		if (stack == focused) setcolor(Selected);
		else setcolor(Occupied);
		wa.border_pixel = color.pixel;
		XChangeWindowAttributes(dpy,stack->win,CWBorderPixel,&wa);
		stack = stack->next;
	}
	if (focused) {
			XRaiseWindow(dpy,focused->win);
			XSetInputFocus(dpy,focused->win,RevertToPointerRoot,CurrentTime);
	}
}
	XFree(info);
	XFlush(dpy);
	return 0;
}

int neighbors(Client *c) {
	prevwin = NULL; nextwin = NULL;
	if (!(c->tags & tagsSel)) return -1;
	{
	Client *stack;
	
	for (stack = clients; stack && stack != c;stack = stack->next)
		if( stack->tags & tagsSel) prevwin = stack;

	if(!prevwin){
		for(nextwin = stack->next; nextwin && nextwin->tags & tagsSel; nextwin = nextwin->next)
		if(nextwin->tags  & tagsSel && nextwin)prevwin=nextwin;
	}
		if( stack->next)nextwin = stack->next;
		if(!nextwin && clients)nextwin=clients;
	}
	return 0;
}


GC setcolor(int col) {
	XAllocNamedColor(dpy,cmap,colors[col],&color,&color);
	XSetForeground(dpy,gc,color.pixel);
	return gc;
}

Client *wintoclient(Window w) {
	Client *c;
	for (c = clients; c && c->win != w; c = c->next);
	return c;
}

Client*  create(Display *dpy, Window win)
{
	if (wintoclient(win))return NULL;

	XWindowAttributes attr;
	XGetWindowAttributes(dpy, win, &attr);
	if(attr.map_state==IsUnmapped ||attr.override_redirect)return NULL;

	Client *c = malloc(sizeof(Client));
	XMapRequestEvent ev;
	ev.display = dpy;
	ev.window = win;
	maprequest((XEvent*)&ev);

return c;
}
// Manage all existing windows
void
getexisting(Display *dpy, Window root)
{
	Window w_;
	Window *childs;
	unsigned int nchilds;
	XQueryTree(dpy, root, &w_, &w_, &childs, &nchilds);

	int i = 0;
	for (i = 0; i < nchilds; i++)
	{
		if (childs[i] == root) continue;
		create(dpy, childs[i]);
	}
	checkrules(NULL);
}
void grabkeys(){
	/* key and mouse bindings */
	unsigned int mods[] = {0, LockMask, Mod2Mask, LockMask|Mod2Mask};
	KeyCode code;
	XUngrabKey(dpy,AnyKey,AnyModifier,root);
	int i,j;
	for (i = 0; i < sizeof(keys)/sizeof(keys[0]); i++)
		if ( (code=XKeysymToKeycode(dpy,keys[i].keysym)) ) for (j = 0; j < 4; j++)
			XGrabKey(dpy,code,keys[i].mod|mods[j],root,True,
					GrabModeAsync,GrabModeAsync);
	for (i = 0; i < sizeof(buttons)/sizeof(buttons[0]); i++)
		if (buttons[i].mod) for (j = 0; j < 4; j++)
			XGrabButton(dpy,buttons[i].button,buttons[i].mod|mods[j],root,True,
					ButtonPressMask,GrabModeAsync,GrabModeAsync,None,None);
}
void run(){
		XEvent ev;
	while (running) {
			XNextEvent(dpy,&ev);
			if (handler[ev.type]) handler[ev.type](&ev);
	}
}

int main(int argc, const char **argv) {
	/* init X */
	if (!(dpy=XOpenDisplay(0x0))) return 1;
	scr = DefaultScreen(dpy);
	sw = DisplayWidth(dpy,scr);
	sh = DisplayHeight(dpy,scr);
	root = DefaultRootWindow(dpy);
	XDefineCursor(dpy,root,XCreateFontCursor(dpy,ttwm_cursor));
	/* gc init */
	cmap = DefaultColormap(dpy,scr);
	XColor color;
	XGCValues val;
	val.font = XLoadFont(dpy,font);
	gc = XCreateGC(dpy,root,GCFont,&val);
	bgc = DefaultGC(dpy,scr);
	XSetForeground(dpy,bgc,color.pixel);
	
	XSetWindowAttributes wa;
	wa.override_redirect = True;
	wa.event_mask = ExposureMask |FocusChangeMask | SubstructureNotifyMask |
			ButtonReleaseMask | PropertyChangeMask | SubstructureRedirectMask |
			StructureNotifyMask;
	XChangeWindowAttributes(dpy,root,CWEventMask,&wa);
	XSelectInput(dpy,root,wa.event_mask);

	getexisting(dpy,root);
	grabkeys();
	draw();
	run();
	XCloseDisplay(dpy);
	return retval;
}
